import ctypes.wintypes

TOKEN_ALL_ACCESS = 0xF01FF

GetCurrentProcess = ctypes.windll.kernel32.GetCurrentProcess
GetCurrentProcess.restype = ctypes.wintypes.HANDLE
OpenProcessToken = ctypes.windll.advapi32.OpenProcessToken
OpenProcessToken.argtypes = (
    ctypes.wintypes.HANDLE,
    ctypes.wintypes.DWORD,
    ctypes.POINTER(ctypes.wintypes.HANDLE),
)
OpenProcessToken.restype = ctypes.wintypes.BOOL



#!python

from __future__ import print_function

import ctypes
from jaraco.windows.error import handle_nonzero_success
from jaraco.windows.api import system
from jaraco.ui.cmdline import Command


def set(value):
    result = system.SystemParametersInfo(
        system.SPI_SETACTIVEWINDOWTRACKING, 0, ctypes.cast(value, ctypes.c_void_p), 0
    )
    handle_nonzero_success(result)


def get():
    value = ctypes.wintypes.BOOL()
    result = system.SystemParametersInfo(
        system.SPI_GETACTIVEWINDOWTRACKING, 0, ctypes.byref(value), 0
    )
    handle_nonzero_success(result)
    return bool(value)


def set_delay(milliseconds):
    result = system.SystemParametersInfo(
        system.SPI_SETACTIVEWNDTRKTIMEOUT,
        0,
        ctypes.cast(milliseconds, ctypes.c_void_p),
        0,
    )
    handle_nonzero_success(result)


def get_delay():
    value = ctypes.wintypes.DWORD()
    result = system.SystemParametersInfo(
        system.SPI_GETACTIVEWNDTRKTIMEOUT, 0, ctypes.byref(value), 0
    )
    handle_nonzero_success(result)
    return int(value.value)


class DelayParam(Command):
    @staticmethod
    def add_arguments(parser):
        parser.add_argument(
            '-d',
            '--delay',
            type=int,
            help="Delay in milliseconds for active window tracking",
        )


class Show(Command):
    @classmethod
    def run(cls, args):
        msg = "xmouse: {enabled} (delay {delay}ms)".format(
            enabled=get(), delay=get_delay()
        )
        print(msg)


class Enable(DelayParam):
    @classmethod
    def run(cls, args):
        print("enabling xmouse")
        set(True)
        args.delay and set_delay(args.delay)


class Disable(DelayParam):
    @classmethod
    def run(cls, args):
        print("disabling xmouse")
        set(False)
        args.delay and set_delay(args.delay)


class Toggle(DelayParam):
    @classmethod
    def run(cls, args):
        value = get()
        print("xmouse: %s -> %s" % (value, not value))
        set(not value)
        args.delay and set_delay(args.delay)


if __name__ == '__main__':
    Command.invoke()



#!/usr/bin/env python

import ctypes


def ensure_unicode(param):
    try:
        param = ctypes.create_unicode_buffer(param)
    except TypeError:
        pass  # just return the param as is
    return param


class Extended(object):
    "Used to add extended capability to structures"

    def __eq__(self, other):
        return memoryview(self) == memoryview(other)

    def __ne__(self, other):
        return memoryview(self) != memoryview(other)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import textwrap

import pytest

import jaraco.windows.clipboard as wc
from jaraco.windows.api import clipboard as api


def test_unicode_clipboard():
    wc.set_unicode_text('foo' * 100)
    assert wc.get_unicode_text() == 'foo' * 100
    wc.set_unicode_text('☃')
    assert wc.get_unicode_text() == '☃'


example_html = textwrap.dedent(
    """
    Version:0.9
    StartHTML:71
    EndHTML:170
    StartFragment:140
    EndFragment:160
    StartSelection:140
    EndSelection:160
    <!DOCTYPE>
    <HTML>
    <HEAD>
    <TITLE> The HTML Clipboard</TITLE>
    <BASE HREF="http://sample/specs">
    </HEAD>
    <BODY>
    <UL>
    <!--StartFragment -->
    <LI> The Fragment </LI>
    <!--EndFragment -->
    </UL>
    </BODY>
    </HTML>
    """
).lstrip()
"""
Example taken from
https://msdn.microsoft.com/en-us/library/windows/desktop/ms649015(v=vs.85).aspx#Description
"""


@pytest.fixture
def sample_html():
    with wc.context():
        wc.SetClipboardData(api.CF_HTML, example_html.encode('utf-8'))


def test_html_paste(sample_html):
    res = wc.get_html()
    assert len(res.html) < len(res.data)
    assert res.headers == dict(
        StartHTML=71,
        Version=0.9,
        EndHTML=170,
        StartFragment=140,
        EndFragment=160,
        StartSelection=140,
        EndSelection=160,
    )


@pytest.mark.xfail(reason="example is buggy")
def test_html_fragment(sample_html):
    snippet = wc.get_html()
    assert snippet.fragment == '<LI> The Fragment </LI>'



"""
API hooks for network stuff.
"""

__all__ = 'AddConnection'

from jaraco.windows.error import WindowsError
from .api import net


def AddConnection(
    remote_name,
    type=net.RESOURCETYPE_ANY,
    local_name=None,
    provider_name=None,
    user=None,
    password=None,
    flags=0,
):
    resource = net.NETRESOURCE(
        type=type,
        remote_name=remote_name,
        local_name=local_name,
        provider_name=provider_name,
        # WNetAddConnection2 ignores the other members of NETRESOURCE
    )

    result = net.WNetAddConnection2(resource, password, user, flags)

    if result != 0:
        raise WindowsError(result)



import os

import pytest

from jaraco.windows import filesystem


def test_islink_on_nonexistent_target():
    assert filesystem.islink('/doesnotexist') is False


def test_is_symlink(tmpdir):
    with tmpdir.as_cwd():
        filesystem.symlink('foobaz', 'foobar')
        assert filesystem.is_symlink('foobar')


def test_FileAttributes():
    attrs = filesystem.FileAttributes.get(__file__)
    assert not attrs.hidden


@pytest.mark.skip('sys.version_info < (3, 2)')
def test_compat_stat():
    """
    Ensure compat stat produces comparable results to
    normal stat on Python 3.2 and later.
    """
    assert filesystem.compat_stat(__file__) == os.stat(__file__)


def test_samefile():
    assert filesystem.samefile(__file__, __file__)
    assert not filesystem.samefile(__file__, filesystem.__file__)



"""
Some routines for retrieving the addresses from the local
network config.
"""

from __future__ import print_function

import itertools
import ctypes


from jaraco.windows.api import errors, inet


def GetAdaptersAddresses():
    size = ctypes.c_ulong()
    res = inet.GetAdaptersAddresses(0, 0, None, None, size)
    if res != errors.ERROR_BUFFER_OVERFLOW:
        raise RuntimeError("Error getting structure length (%d)" % res)
    print(size.value)
    pointer_type = ctypes.POINTER(inet.IP_ADAPTER_ADDRESSES)
    buffer = ctypes.create_string_buffer(size.value)
    struct_p = ctypes.cast(buffer, pointer_type)
    res = inet.GetAdaptersAddresses(0, 0, None, struct_p, size)
    if res != errors.NO_ERROR:
        raise RuntimeError("Error retrieving table (%d)" % res)
    while struct_p:
        yield struct_p.contents
        struct_p = struct_p.contents.next


class AllocatedTable(object):
    """
    Both the interface table and the ip address table use the same
    technique to store arrays of structures of variable length. This
    base class captures the functionality to retrieve and access those
    table entries.

    The subclass needs to define three class attributes:
        method: a callable that takes three arguments - a pointer to
                the structure, the length of the data contained by the
                structure, and a boolean of whether the result should
                be sorted.
        structure: a C structure defininition that describes the table
                format.
        row_structure: a C structure definition that describes the row
                format.
    """

    def __get_table_size(self):
        """
        Retrieve the size of the buffer needed by calling the method
        with a null pointer and length of zero. This should trigger an
        insufficient buffer error and return the size needed for the
        buffer.
        """
        length = ctypes.wintypes.DWORD()
        res = self.method(None, length, False)
        if res != errors.ERROR_INSUFFICIENT_BUFFER:
            raise RuntimeError("Error getting table length (%d)" % res)
        return length.value

    def get_table(self):
        """
        Get the table
        """
        buffer_length = self.__get_table_size()
        returned_buffer_length = ctypes.wintypes.DWORD(buffer_length)
        buffer = ctypes.create_string_buffer(buffer_length)
        pointer_type = ctypes.POINTER(self.structure)
        table_p = ctypes.cast(buffer, pointer_type)
        res = self.method(table_p, returned_buffer_length, False)
        if res != errors.NO_ERROR:
            raise RuntimeError("Error retrieving table (%d)" % res)
        return table_p.contents

    @property
    def entries(self):
        """
        Using the table structure, return the array of entries based
        on the table size.
        """
        table = self.get_table()
        entries_array = self.row_structure * table.num_entries
        pointer_type = ctypes.POINTER(entries_array)
        return ctypes.cast(table.entries, pointer_type).contents


class InterfaceTable(AllocatedTable):
    method = inet.GetIfTable
    structure = inet.MIB_IFTABLE
    row_structure = inet.MIB_IFROW


class AddressTable(AllocatedTable):
    method = inet.GetIpAddrTable
    structure = inet.MIB_IPADDRTABLE
    row_structure = inet.MIB_IPADDRROW


class AddressManager(object):
    @staticmethod
    def hardware_address_to_string(addr):
        hex_bytes = (byte.encode('hex') for byte in addr)
        return ':'.join(hex_bytes)

    def get_host_mac_address_strings(self):
        return (
            self.hardware_address_to_string(addr)
            for addr in self.get_host_mac_addresses()
        )

    def get_host_ip_address_strings(self):
        return itertools.imap(str, self.get_host_ip_addresses())

    def get_host_mac_addresses(self):
        return (entry.physical_address for entry in InterfaceTable().entries)

    def get_host_ip_addresses(self):
        return (entry.address for entry in AddressTable().entries)



from ctypes import CDLL, c_char_p


def get_libc():
    libnames = ('msvcrt', 'libc.so.6')
    for libname in libnames:
        try:
            return CDLL(libname)
        except WindowsError:
            pass
        except OSError:
            pass
    raise RuntimeError("Unable to find a suitable libc (tried %s)" % libnames)


getenv = get_libc().getenv
getenv.restype = c_char_p

# call into your linked module here

print('new value is', getenv('FOO'))



"""
Windows Services support for controlling Windows Services.

Based on http://code.activestate.com
/recipes/115875-controlling-windows-services/
"""

from __future__ import print_function

import sys
import time

import win32api
import win32con
import win32service


class Service(object):
    """
    The Service Class is used for controlling Windows
    services. Just pass the name of the service you wish to control to the
    class instance and go from there. For example, if you want to control
    the Workstation service try this:

        from jaraco.windows import services
        workstation = services.Service("Workstation")
        workstation.start()
        workstation.fetchstatus("running", 10)
        workstation.stop()
        workstation.fetchstatus("stopped")

    Creating an instance of the Service class is done by passing the name of
    the service as it appears in the Management Console or the short name as
    it appears in the registry. Mixed case is ok.
        cvs = services.Service("CVS NT Service 1.11.1.2 (Build 41)")
            or
        cvs = services.Service("cvs")

    If needing remote service control try this:
        cvs = services.Service("cvs", r"\\CVS_SERVER")
            or
        cvs = services.Service("cvs", "\\\\CVS_SERVER")

    The Service Class supports these methods:

        start: Starts service.
        stop: Stops service.
        restart: Stops and restarts service.
        pause: Pauses service (Only if service supports feature).
        resume: Resumes service that has been paused.
        status: Queries current status of service.
        fetchstatus: Continually queries service until requested
            status(STARTING, RUNNING,
            STOPPING & STOPPED) is met or timeout value(in seconds) reached.
            Default timeout value is infinite.
        infotype: Queries service for process type. (Single, shared and/or
            interactive process)
        infoctrl: Queries control information about a running service.
            i.e. Can it be paused, stopped, etc?
        infostartup: Queries service Startup type. (Boot, System,
            Automatic, Manual, Disabled)
        setstartup: Changes/sets Startup type. (Boot, System,
            Automatic, Manual, Disabled)
        getname: Gets the long and short service names used by Windowin32service.
            (Generally used for internal purposes)
    """

    def __init__(self, service, machinename=None, dbname=None):
        self.userv = service
        self.scmhandle = win32service.OpenSCManager(
            machinename, dbname, win32service.SC_MANAGER_ALL_ACCESS
        )
        self.sserv, self.lserv = self.getname()
        if (self.sserv or self.lserv) is None:
            sys.exit()
        self.handle = win32service.OpenService(
            self.scmhandle, self.sserv, win32service.SERVICE_ALL_ACCESS
        )
        self.sccss = "SYSTEM\\CurrentControlSet\\Services\\"

    def start(self):
        win32service.StartService(self.handle, None)

    def stop(self):
        self.stat = win32service.ControlService(
            self.handle, win32service.SERVICE_CONTROL_STOP
        )

    def restart(self):
        self.stop()
        self.fetchstatus("STOPPED")
        self.start()

    def pause(self):
        self.stat = win32service.ControlService(
            self.handle, win32service.SERVICE_CONTROL_PAUSE
        )

    def resume(self):
        self.stat = win32service.ControlService(
            self.handle, win32service.SERVICE_CONTROL_CONTINUE
        )

    def status(self, prn=0):
        self.stat = win32service.QueryServiceStatus(self.handle)
        if self.stat[1] == win32service.SERVICE_STOPPED:
            if prn == 1:
                print("The %s service is stopped." % self.lserv)
            else:
                return "STOPPED"
        elif self.stat[1] == win32service.SERVICE_START_PENDING:
            if prn == 1:
                print("The %s service is starting." % self.lserv)
            else:
                return "STARTING"
        elif self.stat[1] == win32service.SERVICE_STOP_PENDING:
            if prn == 1:
                print("The %s service is stopping." % self.lserv)
            else:
                return "STOPPING"
        elif self.stat[1] == win32service.SERVICE_RUNNING:
            if prn == 1:
                print("The %s service is running." % self.lserv)
            else:
                return "RUNNING"

    def fetchstatus(self, fstatus, timeout=None):
        self.fstatus = fstatus.upper()
        if timeout is not None:
            timeout = int(timeout)
            timeout *= 2

        def to(timeout):
            time.sleep(0.5)
            if timeout is not None:
                if timeout > 1:
                    timeout -= 1
                    return timeout
                else:
                    return "TO"

        if self.fstatus == "STOPPED":
            while 1:
                self.stat = win32service.QueryServiceStatus(self.handle)
                if self.stat[1] == win32service.SERVICE_STOPPED:
                    self.fstate = "STOPPED"
                    break
                else:
                    timeout = to(timeout)
                    if timeout == "TO":
                        return "TIMEDOUT"
                        break
        elif self.fstatus == "STOPPING":
            while 1:
                self.stat = win32service.QueryServiceStatus(self.handle)
                if self.stat[1] == win32service.SERVICE_STOP_PENDING:
                    self.fstate = "STOPPING"
                    break
                else:
                    timeout = to(timeout)
                    if timeout == "TO":
                        return "TIMEDOUT"
                        break
        elif self.fstatus == "RUNNING":
            while 1:
                self.stat = win32service.QueryServiceStatus(self.handle)
                if self.stat[1] == win32service.SERVICE_RUNNING:
                    self.fstate = "RUNNING"
                    break
                else:
                    timeout = to(timeout)
                    if timeout == "TO":
                        return "TIMEDOUT"
                        break
        elif self.fstatus == "STARTING":
            while 1:
                self.stat = win32service.QueryServiceStatus(self.handle)
                if self.stat[1] == win32service.SERVICE_START_PENDING:
                    self.fstate = "STARTING"
                    break
                else:
                    timeout = to(timeout)
                    if timeout == "TO":
                        return "TIMEDOUT"
                        break

    def infotype(self):
        self.stat = win32service.QueryServiceStatus(self.handle)
        if self.stat[0] and win32service.SERVICE_WIN32_OWN_PROCESS:
            print("The %s service runs in its own process." % self.lserv)
        if self.stat[0] and win32service.SERVICE_WIN32_SHARE_PROCESS:
            print("The %s service shares a process with other services." % self.lserv)
        if self.stat[0] and win32service.SERVICE_INTERACTIVE_PROCESS:
            print("The %s service can interact with the desktop." % self.lserv)

    def infoctrl(self):
        self.stat = win32service.QueryServiceStatus(self.handle)
        if self.stat[2] and win32service.SERVICE_ACCEPT_PAUSE_CONTINUE:
            print("The %s service can be paused." % self.lserv)
        if self.stat[2] and win32service.SERVICE_ACCEPT_STOP:
            print("The %s service can be stopped." % self.lserv)
        if self.stat[2] and win32service.SERVICE_ACCEPT_SHUTDOWN:
            print("The %s service can be shutdown." % self.lserv)

    def infostartup(self):
        self.isuphandle = win32api.RegOpenKeyEx(
            win32con.HKEY_LOCAL_MACHINE, self.sccss + self.sserv, 0, win32con.KEY_READ
        )
        self.isuptype = win32api.RegQueryValueEx(self.isuphandle, "Start")[0]
        win32api.RegCloseKey(self.isuphandle)
        if self.isuptype == 0:
            return "boot"
        elif self.isuptype == 1:
            return "system"
        elif self.isuptype == 2:
            return "automatic"
        elif self.isuptype == 3:
            return "manual"
        elif self.isuptype == 4:
            return "disabled"

    @property
    def suptype(self):
        types = 'boot', 'system', 'automatic', 'manual', 'disabled'
        lookup = dict((name, number) for number, name in enumerate(types))
        return lookup[self.startuptype]

    def setstartup(self, startuptype):
        self.startuptype = startuptype.lower()
        self.snc = win32service.SERVICE_NO_CHANGE
        win32service.ChangeServiceConfig(
            self.handle,
            self.snc,
            self.suptype,
            self.snc,
            None,
            None,
            0,
            None,
            None,
            None,
            self.lserv,
        )

    def getname(self):
        self.snames = win32service.EnumServicesStatus(self.scmhandle)
        for i in self.snames:
            if i[0].lower() == self.userv.lower():
                return i[0], i[1]
                break
            if i[1].lower() == self.userv.lower():
                return i[0], i[1]
                break
        print("Error: The %s service doesn't seem to exist." % self.userv)
        return None, None



import ctypes.wintypes

# Clipboard Formats
CF_TEXT = 1
CF_BITMAP = 2
CF_METAFILEPICT = 3
CF_SYLK = 4
CF_DIF = 5
CF_TIFF = 6
CF_OEMTEXT = 7
CF_DIB = 8
CF_PALETTE = 9
CF_PENDATA = 10
CF_RIFF = 11
CF_WAVE = 12
CF_UNICODETEXT = 13
CF_ENHMETAFILE = 14
CF_HDROP = 15
CF_LOCALE = 16
CF_DIBV5 = 17
CF_MAX = 18
CF_OWNERDISPLAY = 0x0080
CF_DSPTEXT = 0x0081
CF_DSPBITMAP = 0x0082
CF_DSPMETAFILEPICT = 0x0083
CF_DSPENHMETAFILE = 0x008E
CF_PRIVATEFIRST = 0x0200
CF_PRIVATELAST = 0x02FF
CF_GDIOBJFIRST = 0x0300
CF_GDIOBJLAST = 0x03FF

RegisterClipboardFormat = ctypes.windll.user32.RegisterClipboardFormatW
RegisterClipboardFormat.argtypes = (ctypes.wintypes.LPWSTR,)
RegisterClipboardFormat.restype = ctypes.wintypes.UINT
CF_HTML = RegisterClipboardFormat('HTML Format')

EnumClipboardFormats = ctypes.windll.user32.EnumClipboardFormats
EnumClipboardFormats.argtypes = (ctypes.wintypes.UINT,)
EnumClipboardFormats.restype = ctypes.wintypes.UINT

GetClipboardData = ctypes.windll.user32.GetClipboardData
GetClipboardData.argtypes = (ctypes.wintypes.UINT,)
GetClipboardData.restype = ctypes.wintypes.HANDLE

SetClipboardData = ctypes.windll.user32.SetClipboardData
SetClipboardData.argtypes = ctypes.wintypes.UINT, ctypes.wintypes.HANDLE
SetClipboardData.restype = ctypes.wintypes.HANDLE

OpenClipboard = ctypes.windll.user32.OpenClipboard
OpenClipboard.argtypes = (ctypes.wintypes.HANDLE,)
OpenClipboard.restype = ctypes.wintypes.BOOL

ctypes.windll.user32.CloseClipboard.restype = ctypes.wintypes.BOOL



#!/usr/bin/env python

import operator
import ctypes
import datetime
from ctypes.wintypes import WORD, WCHAR, BOOL, LONG

from jaraco.windows.util import Extended
from jaraco.collections import RangeMap


class AnyDict(object):
    "A dictionary that returns the same value regardless of key"

    def __init__(self, value):
        self.value = value

    def __getitem__(self, key):
        return self.value


class SYSTEMTIME(Extended, ctypes.Structure):
    _fields_ = [
        ('year', WORD),
        ('month', WORD),
        ('day_of_week', WORD),
        ('day', WORD),
        ('hour', WORD),
        ('minute', WORD),
        ('second', WORD),
        ('millisecond', WORD),
    ]


class REG_TZI_FORMAT(Extended, ctypes.Structure):
    _fields_ = [
        ('bias', LONG),
        ('standard_bias', LONG),
        ('daylight_bias', LONG),
        ('standard_start', SYSTEMTIME),
        ('daylight_start', SYSTEMTIME),
    ]


class TIME_ZONE_INFORMATION(Extended, ctypes.Structure):
    _fields_ = [
        ('bias', LONG),
        ('standard_name', WCHAR * 32),
        ('standard_start', SYSTEMTIME),
        ('standard_bias', LONG),
        ('daylight_name', WCHAR * 32),
        ('daylight_start', SYSTEMTIME),
        ('daylight_bias', LONG),
    ]


class DYNAMIC_TIME_ZONE_INFORMATION(TIME_ZONE_INFORMATION):
    """
    Because the structure of the DYNAMIC_TIME_ZONE_INFORMATION extends
    the structure of the TIME_ZONE_INFORMATION, this structure
    can be used as a drop-in replacement for calls where the
    structure is passed by reference.

    For example,
    dynamic_tzi = DYNAMIC_TIME_ZONE_INFORMATION()
    ctypes.windll.kernel32.GetTimeZoneInformation(ctypes.byref(dynamic_tzi))

    (although the key_name and dynamic_daylight_time_disabled flags will be
    set to the default (null)).

    >>> isinstance(DYNAMIC_TIME_ZONE_INFORMATION(), TIME_ZONE_INFORMATION)
    True


    """

    _fields_ = [
        # ctypes automatically includes the fields from the parent
        ('key_name', WCHAR * 128),
        ('dynamic_daylight_time_disabled', BOOL),
    ]

    def __init__(self, *args, **kwargs):
        """Allow initialization from args from both this class and
        its superclass.  Default ctypes implementation seems to
        assume that this class is only initialized with its own
        _fields_ (for non-keyword-args)."""
        super_self = super(DYNAMIC_TIME_ZONE_INFORMATION, self)
        super_fields = super_self._fields_
        super_args = args[: len(super_fields)]
        self_args = args[len(super_fields) :]
        # convert the super args to keyword args so they're also handled
        for field, arg in zip(super_fields, super_args):
            field_name, spec = field
            kwargs[field_name] = arg
        super(DYNAMIC_TIME_ZONE_INFORMATION, self).__init__(*self_args, **kwargs)


class Info(DYNAMIC_TIME_ZONE_INFORMATION):
    """
    A time zone definition class based on the win32
    DYNAMIC_TIME_ZONE_INFORMATION structure.

    Describes a bias against UTC (bias), and two dates at which a separate
    additional bias applies (standard_bias and daylight_bias).
    """

    def field_names(self):
        return map(operator.itemgetter(0), self._fields_)

    def __init__(self, *args, **kwargs):
        """
        Try to construct a timezone.Info from
        a) [DYNAMIC_]TIME_ZONE_INFORMATION args
        b) another Info
        c) a REG_TZI_FORMAT
        d) a byte structure
        """
        funcs = (
            super(Info, self).__init__,
            self.__init_from_other,
            self.__init_from_reg_tzi,
            self.__init_from_bytes,
        )
        for func in funcs:
            try:
                func(*args, **kwargs)
                return
            except TypeError:
                pass
        raise TypeError("Invalid arguments for %s" % self.__class__)

    def __init_from_bytes(self, bytes, **kwargs):
        reg_tzi = REG_TZI_FORMAT()
        # todo: use buffer API in Python 3
        buffer = memoryview(bytes)
        ctypes.memmove(ctypes.addressof(reg_tzi), buffer, len(buffer))
        self.__init_from_reg_tzi(self, reg_tzi, **kwargs)

    def __init_from_reg_tzi(self, reg_tzi, **kwargs):
        if not isinstance(reg_tzi, REG_TZI_FORMAT):
            raise TypeError("Not a REG_TZI_FORMAT")
        for field_name, type in reg_tzi._fields_:
            setattr(self, field_name, getattr(reg_tzi, field_name))
        for name, value in kwargs.items():
            setattr(self, name, value)

    def __init_from_other(self, other):
        if not isinstance(other, TIME_ZONE_INFORMATION):
            raise TypeError("Not a TIME_ZONE_INFORMATION")
        for name in other.field_names():
            # explicitly get the value from the underlying structure
            value = super(Info, other).__getattribute__(other, name)
            setattr(self, name, value)
        # consider instead of the loop above just copying the memory directly
        # size = max(ctypes.sizeof(DYNAMIC_TIME_ZONE_INFO), ctypes.sizeof(other))
        # ctypes.memmove(ctypes.addressof(self), other, size)

    def __getattribute__(self, attr):
        value = super(Info, self).__getattribute__(attr)

        def make_minute_timedelta(m):
            datetime.timedelta(minutes=m)

        if 'bias' in attr:
            value = make_minute_timedelta(value)
        return value

    @classmethod
    def current(class_):
        "Windows Platform SDK GetTimeZoneInformation"
        tzi = class_()
        kernel32 = ctypes.windll.kernel32
        getter = kernel32.GetTimeZoneInformation
        getter = getattr(kernel32, 'GetDynamicTimeZoneInformation', getter)
        code = getter(ctypes.byref(tzi))
        return code, tzi

    def set(self):
        kernel32 = ctypes.windll.kernel32
        setter = kernel32.SetTimeZoneInformation
        setter = getattr(kernel32, 'SetDynamicTimeZoneInformation', setter)
        return setter(ctypes.byref(self))

    def copy(self):
        return self.__class__(self)

    def locate_daylight_start(self, year):
        info = self.get_info_for_year(year)
        return self._locate_day(year, info.daylight_start)

    def locate_standard_start(self, year):
        info = self.get_info_for_year(year)
        return self._locate_day(year, info.standard_start)

    def get_info_for_year(self, year):
        return self.dynamic_info[year]

    @property
    def dynamic_info(self):
        "Return a map that for a given year will return the correct Info"
        if self.key_name:
            dyn_key = self.get_key().subkey('Dynamic DST')
            del dyn_key['FirstEntry']
            del dyn_key['LastEntry']
            years = map(int, dyn_key.keys())
            values = map(Info, dyn_key.values())
            # create a range mapping that searches by descending year and matches
            # if the target year is greater or equal.
            return RangeMap(zip(years, values), RangeMap.descending, operator.ge)
        else:
            return AnyDict(self)

    @staticmethod
    def _locate_day(year, cutoff):
        """
        Takes a SYSTEMTIME object, such as retrieved from a TIME_ZONE_INFORMATION
        structure or call to GetTimeZoneInformation and interprets
        it based on the given
        year to identify the actual day.

        This method is necessary because the SYSTEMTIME structure
        refers to a day by its
        day of the week and week of the month (e.g. 4th saturday in March).

        >>> SATURDAY = 6
        >>> MARCH = 3
        >>> st = SYSTEMTIME(2000, MARCH, SATURDAY, 4, 0, 0, 0, 0)

        # according to my calendar, the 4th Saturday in March in 2009 was the 28th
        >>> expected_date = datetime.datetime(2009, 3, 28)
        >>> Info._locate_day(2009, st) == expected_date
        True
        """
        # MS stores Sunday as 0, Python datetime stores Monday as zero
        target_weekday = (cutoff.day_of_week + 6) % 7
        # For SYSTEMTIMEs relating to time zone inforamtion, cutoff.day
        #  is the week of the month
        week_of_month = cutoff.day
        # so the following is the first day of that week
        day = (week_of_month - 1) * 7 + 1
        result = datetime.datetime(
            year,
            cutoff.month,
            day,
            cutoff.hour,
            cutoff.minute,
            cutoff.second,
            cutoff.millisecond,
        )
        # now the result is the correct week, but not necessarily
        # the correct day of the week
        days_to_go = (target_weekday - result.weekday()) % 7
        result += datetime.timedelta(days_to_go)
        # if we selected a day in the month following the target month,
        #  move back a week or two.
        # This is necessary because Microsoft defines the fifth week in a month
        #  to be the last week in a month and adding the time delta might have
        #  pushed the result into the next month.
        while result.month == cutoff.month + 1:
            result -= datetime.timedelta(weeks=1)
        return result



import ctypes.wintypes

import six

from .error import handle_nonzero_success
from .api import memory


class MemoryMap(object):
    """
    A memory map object which can have security attributes overridden.
    """

    def __init__(self, name, length, security_attributes=None):
        self.name = name
        self.length = length
        self.security_attributes = security_attributes
        self.pos = 0

    def __enter__(self):
        p_SA = (
            ctypes.byref(self.security_attributes) if self.security_attributes else None
        )
        INVALID_HANDLE_VALUE = -1
        PAGE_READWRITE = 0x4
        FILE_MAP_WRITE = 0x2
        filemap = ctypes.windll.kernel32.CreateFileMappingW(
            INVALID_HANDLE_VALUE,
            p_SA,
            PAGE_READWRITE,
            0,
            self.length,
            six.text_type(self.name),
        )
        handle_nonzero_success(filemap)
        if filemap == INVALID_HANDLE_VALUE:
            raise Exception("Failed to create file mapping")
        self.filemap = filemap
        self.view = memory.MapViewOfFile(filemap, FILE_MAP_WRITE, 0, 0, 0)
        return self

    def seek(self, pos):
        self.pos = pos

    def write(self, msg):
        assert isinstance(msg, bytes)
        n = len(msg)
        if self.pos + n >= self.length:  # A little safety.
            raise ValueError("Refusing to write %d bytes" % n)
        dest = self.view + self.pos
        length = ctypes.c_size_t(n)
        ctypes.windll.kernel32.RtlMoveMemory(dest, msg, length)
        self.pos += n

    def read(self, n):
        """
        Read n bytes from mapped view.
        """
        out = ctypes.create_string_buffer(n)
        source = self.view + self.pos
        length = ctypes.c_size_t(n)
        ctypes.windll.kernel32.RtlMoveMemory(out, source, length)
        self.pos += n
        return out.raw

    def __exit__(self, exc_type, exc_val, tb):
        ctypes.windll.kernel32.UnmapViewOfFile(self.view)
        ctypes.windll.kernel32.CloseHandle(self.filemap)



import platform
import os

import pytest


collect_ignore = ['bugs', 'incubator']


try:
    __import__('win32api')
except ImportError:
    # pywin32 isn't available, so avoid import errors
    collect_ignore += ['jaraco/windows/eventlog.py', 'jaraco/windows/services.py']


if platform.system() != 'Windows':
    """
    Ignore everything but 'test_root' on non-Windows systems
    """
    collect_ignore = [
        os.path.join(root, filename)
        for root, dirs, files in os.walk('.')
        for filename in files
        if filename.endswith('.py') and not filename.startswith('test_root')
    ]


@pytest.fixture
def tmpdir_as_cwd(tmpdir):
    with tmpdir.as_cwd():
        yield tmpdir



#!/usr/bin/env python

"""
jaraco.windows.message

Windows Messaging support
"""

import ctypes
from ctypes.wintypes import HWND, UINT, WPARAM, LPARAM, DWORD, LPVOID

import six

LRESULT = LPARAM


class LPARAM_wstr(LPARAM):
    """
    A special instance of LPARAM that can be constructed from a string
    instance (for functions such as SendMessage, whose LPARAM may point to
    a unicode string).
    """

    @classmethod
    def from_param(cls, param):
        if isinstance(param, six.string_types):
            return LPVOID.from_param(six.text_type(param))
        return LPARAM.from_param(param)


SendMessage = ctypes.windll.user32.SendMessageW
SendMessage.argtypes = (HWND, UINT, WPARAM, LPARAM_wstr)
SendMessage.restype = LRESULT

HWND_BROADCAST = 0xFFFF
WM_SETTINGCHANGE = 0x1A

# constants from http://msdn.microsoft.com
# /en-us/library/ms644952%28v=vs.85%29.aspx
SMTO_ABORTIFHUNG = 0x02
SMTO_BLOCK = 0x01
SMTO_NORMAL = 0x00
SMTO_NOTIMEOUTIFNOTHUNG = 0x08
SMTO_ERRORONEXIT = 0x20

SendMessageTimeout = ctypes.windll.user32.SendMessageTimeoutW
SendMessageTimeout.argtypes = SendMessage.argtypes + (UINT, UINT, ctypes.POINTER(DWORD))
SendMessageTimeout.restype = LRESULT


def unicode_as_lparam(source):
    pointer = ctypes.cast(ctypes.c_wchar_p(source), ctypes.c_void_p)
    return LPARAM(pointer.value)



"""
Support for Credential Vault
"""

import ctypes
from ctypes.wintypes import DWORD, LPCWSTR, BOOL, LPWSTR, FILETIME


try:
    from ctypes.wintypes import LPBYTE
except ImportError:
    LPBYTE = ctypes.POINTER(ctypes.wintypes.BYTE)  # type: ignore


class CredentialAttribute(ctypes.Structure):
    _fields_ = []  # type: ignore


class Credential(ctypes.Structure):
    _fields_ = [
        ('flags', DWORD),
        ('type', DWORD),
        ('target_name', LPWSTR),
        ('comment', LPWSTR),
        ('last_written', FILETIME),
        ('credential_blob_size', DWORD),
        ('credential_blob', LPBYTE),
        ('persist', DWORD),
        ('attribute_count', DWORD),
        ('attributes', ctypes.POINTER(CredentialAttribute)),
        ('target_alias', LPWSTR),
        ('user_name', LPWSTR),
    ]

    def __del__(self):
        ctypes.windll.advapi32.CredFree(ctypes.byref(self))


PCREDENTIAL = ctypes.POINTER(Credential)

CredRead = ctypes.windll.advapi32.CredReadW
CredRead.argtypes = (
    LPCWSTR,  # TargetName
    DWORD,  # Type
    DWORD,  # Flags
    ctypes.POINTER(PCREDENTIAL),  # Credential
)
CredRead.restype = BOOL

CredWrite = ctypes.windll.advapi32.CredWriteW
CredWrite.argtypes = (PCREDENTIAL, DWORD)  # Credential  # Flags
CredWrite.restype = BOOL

CredDelete = ctypes.windll.advapi32.CredDeleteW
CredDelete.argtypes = (LPCWSTR, DWORD, DWORD)  # TargetName  # Type  # Flags
CredDelete.restype = BOOL



#!/usr/bin/env python

import ctypes
from jaraco.windows.util import ensure_unicode


def MessageBox(text, caption=None, handle=None, type=None):
    text, caption = map(ensure_unicode, (text, caption))
    ctypes.windll.user32.MessageBoxW(handle, text, caption, type)



#!/usr/bin/env python

import sys

import ctypes
import ctypes.wintypes

import six

builtins = six.moves.builtins


__import__('jaraco.windows.api.memory')


def format_system_message(errno):
    """
    Call FormatMessage with a system error number to retrieve
    the descriptive error message.
    """
    # first some flags used by FormatMessageW
    ALLOCATE_BUFFER = 0x100
    FROM_SYSTEM = 0x1000

    # Let FormatMessageW allocate the buffer (we'll free it below)
    # Also, let it know we want a system error message.
    flags = ALLOCATE_BUFFER | FROM_SYSTEM
    source = None
    message_id = errno
    language_id = 0
    result_buffer = ctypes.wintypes.LPWSTR()
    buffer_size = 0
    arguments = None
    bytes = ctypes.windll.kernel32.FormatMessageW(
        flags,
        source,
        message_id,
        language_id,
        ctypes.byref(result_buffer),
        buffer_size,
        arguments,
    )
    # note the following will cause an infinite loop if GetLastError
    #  repeatedly returns an error that cannot be formatted, although
    #  this should not happen.
    handle_nonzero_success(bytes)
    message = result_buffer.value
    ctypes.windll.kernel32.LocalFree(result_buffer)
    return message


class WindowsError(builtins.WindowsError):
    """
    More info about errors at
    http://msdn.microsoft.com/en-us/library/ms681381(VS.85).aspx
    """

    def __init__(self, value=None):
        if value is None:
            value = ctypes.windll.kernel32.GetLastError()
        strerror = format_system_message(value)
        if sys.version_info > (3, 3):
            args = 0, strerror, None, value
        else:
            args = value, strerror
        super(WindowsError, self).__init__(*args)

    @property
    def message(self):
        return self.strerror

    @property
    def code(self):
        return self.winerror

    def __str__(self):
        return self.message

    def __repr__(self):
        return '{self.__class__.__name__}({self.winerror})'.format(**vars())


def handle_nonzero_success(result):
    if result == 0:
        raise WindowsError()



import ctypes.wintypes


class LUID(ctypes.Structure):
    _fields_ = [
        ('low_part', ctypes.wintypes.DWORD),
        ('high_part', ctypes.wintypes.LONG),
    ]

    def __eq__(self, other):
        return self.high_part == other.high_part and self.low_part == other.low_part

    def __ne__(self, other):
        return not (self == other)


LookupPrivilegeValue = ctypes.windll.advapi32.LookupPrivilegeValueW
LookupPrivilegeValue.argtypes = (
    ctypes.wintypes.LPWSTR,  # system name
    ctypes.wintypes.LPWSTR,  # name
    ctypes.POINTER(LUID),
)
LookupPrivilegeValue.restype = ctypes.wintypes.BOOL


class TOKEN_INFORMATION_CLASS:
    TokenUser = 1
    TokenGroups = 2
    TokenPrivileges = 3
    # ... see http://msdn.microsoft.com/en-us/library/aa379626%28VS.85%29.aspx


SE_PRIVILEGE_ENABLED_BY_DEFAULT = 0x00000001
SE_PRIVILEGE_ENABLED = 0x00000002
SE_PRIVILEGE_REMOVED = 0x00000004
SE_PRIVILEGE_USED_FOR_ACCESS = 0x80000000


class LUID_AND_ATTRIBUTES(ctypes.Structure):
    _fields_ = [('LUID', LUID), ('attributes', ctypes.wintypes.DWORD)]

    def is_enabled(self):
        return bool(self.attributes & SE_PRIVILEGE_ENABLED)

    def enable(self):
        self.attributes |= SE_PRIVILEGE_ENABLED

    def get_name(self):
        size = ctypes.wintypes.DWORD(10240)
        buf = ctypes.create_unicode_buffer(size.value)
        res = LookupPrivilegeName(None, self.LUID, buf, size)
        if res == 0:
            raise RuntimeError
        return buf[: size.value]

    def __str__(self):
        res = self.get_name()
        if self.is_enabled():
            res += ' (enabled)'
        return res


LookupPrivilegeName = ctypes.windll.advapi32.LookupPrivilegeNameW
LookupPrivilegeName.argtypes = (
    ctypes.wintypes.LPWSTR,  # lpSystemName
    ctypes.POINTER(LUID),  # lpLuid
    ctypes.wintypes.LPWSTR,  # lpName
    ctypes.POINTER(ctypes.wintypes.DWORD),  # cchName
)
LookupPrivilegeName.restype = ctypes.wintypes.BOOL


class TOKEN_PRIVILEGES(ctypes.Structure):
    _fields_ = [
        ('count', ctypes.wintypes.DWORD),
        ('privileges', LUID_AND_ATTRIBUTES * 0),
    ]

    def get_array(self):
        array_type = LUID_AND_ATTRIBUTES * self.count
        privileges = ctypes.cast(self.privileges, ctypes.POINTER(array_type)).contents
        return privileges

    def __iter__(self):
        return iter(self.get_array())


PTOKEN_PRIVILEGES = ctypes.POINTER(TOKEN_PRIVILEGES)

GetTokenInformation = ctypes.windll.advapi32.GetTokenInformation
GetTokenInformation.argtypes = [
    ctypes.wintypes.HANDLE,  # TokenHandle
    ctypes.c_uint,  # TOKEN_INFORMATION_CLASS value
    ctypes.c_void_p,  # TokenInformation
    ctypes.wintypes.DWORD,  # TokenInformationLength
    ctypes.POINTER(ctypes.wintypes.DWORD),  # ReturnLength
]
GetTokenInformation.restype = ctypes.wintypes.BOOL

# http://msdn.microsoft.com/en-us/library/aa375202%28VS.85%29.aspx
AdjustTokenPrivileges = ctypes.windll.advapi32.AdjustTokenPrivileges
AdjustTokenPrivileges.restype = ctypes.wintypes.BOOL
AdjustTokenPrivileges.argtypes = [
    ctypes.wintypes.HANDLE,  # TokenHandle
    ctypes.wintypes.BOOL,  # DisableAllPrivileges
    PTOKEN_PRIVILEGES,  # NewState (optional)
    ctypes.wintypes.DWORD,  # BufferLength of PreviousState
    PTOKEN_PRIVILEGES,  # PreviousState (out, optional)
    ctypes.POINTER(ctypes.wintypes.DWORD),  # ReturnLength
]



import ctypes.wintypes

BOOL = ctypes.wintypes.BOOL


class SHELLSTATE(ctypes.Structure):
    _fields_ = [
        ('show_all_objects', BOOL, 1),
        ('show_extensions', BOOL, 1),
        ('no_confirm_recycle', BOOL, 1),
        ('show_sys_files', BOOL, 1),
        ('show_comp_color', BOOL, 1),
        ('double_click_in_web_view', BOOL, 1),
        ('desktop_HTML', BOOL, 1),
        ('win95_classic', BOOL, 1),
        ('dont_pretty_path', BOOL, 1),
        ('show_attrib_col', BOOL, 1),
        ('map_network_drive_button', BOOL, 1),
        ('show_info_tip', BOOL, 1),
        ('hide_icons', BOOL, 1),
        ('web_view', BOOL, 1),
        ('filter', BOOL, 1),
        ('show_super_hidden', BOOL, 1),
        ('no_net_crawling', BOOL, 1),
        ('win95_unused', ctypes.wintypes.DWORD),
        ('param_sort', ctypes.wintypes.LONG),
        ('sort_direction', ctypes.c_int),
        ('version', ctypes.wintypes.UINT),
        ('not_used', ctypes.wintypes.UINT),
        ('sep_process', BOOL, 1),
        ('start_panel_on', BOOL, 1),
        ('show_start_page', BOOL, 1),
        ('auto_check_select', BOOL, 1),
        ('icons_only', BOOL, 1),
        ('show_type_overlay', BOOL, 1),
        ('spare_flags', ctypes.wintypes.UINT, 13),
    ]


SSF_SHOWALLOBJECTS = 0x00000001
"The fShowAllObjects member is being requested."

SSF_SHOWEXTENSIONS = 0x00000002
"The fShowExtensions member is being requested."

SSF_HIDDENFILEEXTS = 0x00000004
"Not used."

SSF_SERVERADMINUI = 0x00000004
"Not used."

SSF_SHOWCOMPCOLOR = 0x00000008
"The fShowCompColor member is being requested."

SSF_SORTCOLUMNS = 0x00000010
"The lParamSort and iSortDirection members are being requested."

SSF_SHOWSYSFILES = 0x00000020
"The fShowSysFiles member is being requested."

SSF_DOUBLECLICKINWEBVIEW = 0x00000080
"The fDoubleClickInWebView member is being requested."

SSF_SHOWATTRIBCOL = 0x00000100
"The fShowAttribCol member is being requested. (Windows Vista: Not used.)"

SSF_DESKTOPHTML = 0x00000200
"""
The fDesktopHTML member is being requested. Set is not available.
Instead, for versions of Microsoft Windows prior to Windows XP,
enable Desktop HTML by IActiveDesktop. The use of IActiveDesktop
for this purpose, however, is not recommended for Windows XP and
later versions of Windows, and is deprecated in Windows Vista.
"""

SSF_WIN95CLASSIC = 0x00000400
"The fWin95Classic member is being requested."

SSF_DONTPRETTYPATH = 0x00000800
"The fDontPrettyPath member is being requested."

SSF_MAPNETDRVBUTTON = 0x00001000
"The fMapNetDrvBtn member is being requested."

SSF_SHOWINFOTIP = 0x00002000
"The fShowInfoTip member is being requested."

SSF_HIDEICONS = 0x00004000
"The fHideIcons member is being requested."

SSF_NOCONFIRMRECYCLE = 0x00008000
"The fNoConfirmRecycle member is being requested."

SSF_FILTER = 0x00010000
"The fFilter member is being requested. (Windows Vista: Not used.)"

SSF_WEBVIEW = 0x00020000
"The fWebView member is being requested."

SSF_SHOWSUPERHIDDEN = 0x00040000
"The fShowSuperHidden member is being requested."

SSF_SEPPROCESS = 0x00080000
"The fSepProcess member is being requested."

SSF_NONETCRAWLING = 0x00100000
"Windows XP and later. The fNoNetCrawling member is being requested."

SSF_STARTPANELON = 0x00200000
"Windows XP and later. The fStartPanelOn member is being requested."

SSF_SHOWSTARTPAGE = 0x00400000
"Not used."

SSF_AUTOCHECKSELECT = 0x00800000
"Windows Vista and later. The fAutoCheckSelect member is being requested."

SSF_ICONSONLY = 0x01000000
"Windows Vista and later. The fIconsOnly member is being requested."

SSF_SHOWTYPEOVERLAY = 0x02000000
"Windows Vista and later. The fShowTypeOverlay member is being requested."


SHGetSetSettings = ctypes.windll.shell32.SHGetSetSettings
SHGetSetSettings.argtypes = [
    ctypes.POINTER(SHELLSTATE),
    ctypes.wintypes.DWORD,
    ctypes.wintypes.BOOL,  # get or set (True: set)
]
SHGetSetSettings.restype = None



from .api import shell


def get_recycle_bin_confirm():
    settings = shell.SHELLSTATE()
    shell.SHGetSetSettings(settings, shell.SSF_NOCONFIRMRECYCLE, False)
    return not settings.no_confirm_recycle


def set_recycle_bin_confirm(confirm=False):
    settings = shell.SHELLSTATE()
    settings.no_confirm_recycle = not confirm
    shell.SHGetSetSettings(settings, shell.SSF_NOCONFIRMRECYCLE, True)
    # cross fingers and hope it worked



import ctypes

import jaraco.windows.api.credential as api
from . import error

CRED_TYPE_GENERIC = 1


def CredDelete(TargetName, Type, Flags=0):
    error.handle_nonzero_success(api.CredDelete(TargetName, Type, Flags))


def CredRead(TargetName, Type, Flags=0):
    cred_pointer = api.PCREDENTIAL()
    res = api.CredRead(TargetName, Type, Flags, ctypes.byref(cred_pointer))
    error.handle_nonzero_success(res)
    return cred_pointer.contents


def CredWrite(Credential, Flags=0):
    res = api.CredWrite(Credential, Flags)
    error.handle_nonzero_success(res)



import os
import sys

try:
    from jaraco.windows.filesystem import symlink
except ImportError:
    # a dirty reimplementation of symlink from jaraco.windows
    from ctypes import windll
    from ctypes.wintypes import LPWSTR, DWORD, BOOLEAN

    CreateSymbolicLink = windll.kernel32.CreateSymbolicLinkW
    CreateSymbolicLink.argtypes = (LPWSTR, LPWSTR, DWORD)
    CreateSymbolicLink.restype = BOOLEAN

    def symlink(link, target, target_is_directory=False):
        """
        An implementation of os.symlink for Windows (Vista and greater)
        """
        target_is_directory = target_is_directory or os.path.isdir(target)
        CreateSymbolicLink(link, target, target_is_directory)


assert sys.platform in ('win32',)
os.makedirs(r'.\foo')
assert os.path.isdir(r'.\foo')

symlink(r'.\foo_sym', r'.\foo')
assert os.path.isdir(r'.\foo_sym')
assert os.path.islink(r'.\foo_sym')  # fails



# reported at http://social.msdn.microsoft.com/Forums/en-US/wsk/thread/f43c2faf-3df3-4f11-9f5e-1a9101753f93
from win32wnet import WNetAddConnection2, NETRESOURCE

resource = NETRESOURCE()
resource.lpRemoteName = r'\\aoshi\users'
username = 'jaraco'
res = WNetAddConnection2(resource, UserName=username)
print('first result is', res)
res = WNetAddConnection2(resource, UserName=username)
print('second result is', res)

"""
Output is:

first result is None
Traceback (most recent call last):
  File ".\wnetaddconnection2-error-on-64-bit.py", line 7, in <module>
    res = WNetAddConnection2(resource, UserName=username)
pywintypes.error: (1219, 'WNetAddConnection2', 'Multiple connections to a server or shared resource by the same user, using more than one user name, are not allowed. Disconnect all previous connections to the server or shared resource and try again.')
"""
